#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <alsa/asoundlib.h>

#include "usb.h"

int quit=0;

void signalHandler(int signal, siginfo_t *info, void *object)
{
 switch(signal)
 {
  case SIGPIPE:
  {
   break;
  }
  default:
  {
   quit=1;
   break;
  }
 }
}



int main(int argc, char **argv)
{
 /*
  * Init
  */

 int r;


 //setup signal handler
 struct sigaction sa;
 sigemptyset(&sa.sa_mask);
 sa.sa_flags=0;
 sa.sa_sigaction=signalHandler;

 if(
    sigaction(SIGPIPE, &sa, 0)<0||
    sigaction(SIGCHLD, &sa, 0)<0||
    sigaction(SIGTERM, &sa, 0)<0||
    sigaction(SIGQUIT, &sa, 0)<0||
    sigaction(SIGABRT, &sa, 0)<0||
    sigaction(SIGSEGV, &sa, 0)<0||
    sigaction(SIGALRM, &sa, 0)<0||
    sigaction(SIGINT, &sa, 0)<0||
    sigaction(SIGHUP, &sa, 0)<0
 )
  return 1;

 //setup USB device
 struct usb_device dev;
 dev.vendor=0x0582; dev.product=0x0150; //Roland / TD-15

 if(usb_open(&dev, 3)!=0)
  return 1;

 //setup usb transfer object
 struct usb_transfer transfer;
 transfer.data=NULL; //ensure data isn't freed the first time


 //setup MIDI device 
 int skipOffs=0;
 snd_rawmidi_t *midiOutputHandle=NULL;
 register int err;

 if(argc>1)
 {
  if(strcmp(argv[1], "--help")==0||strcmp(argv[1], "-h")==0)
  {
   printf("\n%s [midi device] [skipOffs]\n\n", argv[0]);
   printf("Optional paramaters:\n");
   printf("- midi device:  Device which should receive midi signals (e.g. hw:0,0)\n");
   printf("- skipOffs:     Skip midi off signals\n\n");
   return 0;
  }

  if((r=snd_rawmidi_open(NULL, &midiOutputHandle, argv[1], 0))<0)
  {
   printf("Can't open MIDI input: %s\n", snd_strerror(r));
   return 1;
  }

  if(argc>2&&strcmp(argv[2], "skipOffs")==0)
   skipOffs=1;
 }

 /*
  * Main
  */


 /* init device */

 usb_transfer_reset(&transfer, 18, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 9, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 167, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 64, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 18, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 255, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 255, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x300;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 255, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x302;
 transfer.index=0x0409;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 18, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 80, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 167, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x01;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x02;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x02;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_INTERFACE;
 transfer.request=LIBUSB_REQUEST_SET_INTERFACE;
 transfer.value=0x00;
 transfer.index=0x01;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }


 /* ready for data */

 while(!quit)
 {
  usb_transfer_reset(&transfer, 512, NULL);
  transfer.recipient=LIBUSB_RECIPIENT_ENDPOINT;
  transfer.endpoint=LIBUSB_ENDPOINT_IN;

  if(usb_bulk(&dev, &transfer)!=0)
  {
   usb_close(&dev);
   return 2;
  }

  if(transfer.length) //data
  {
   printf("Data: ");
   for(r=0;r<transfer.length;r++)
    printf("%.2x ", transfer.data[r]);

   printf("\n");
   fflush(stdout);
   
   if(transfer.length==4) //midi
   {
    if(skipOffs&&transfer.data[0]!=9) //off signals
     continue;

    //write data to midi
    if(midiOutputHandle)
     if((r=snd_rawmidi_write(midiOutputHandle, transfer.data, transfer.length))<0)
      printf("Can't write to MIDI device: %s\n", snd_strerror(r));
   }
  }
 }


 /* shutdown device */

 usb_transfer_reset(&transfer, 18, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x100;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 167, NULL);
 transfer.endpoint=LIBUSB_ENDPOINT_IN;
 transfer.request=LIBUSB_REQUEST_GET_DESCRIPTOR;
 transfer.value=0x200;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }

 usb_transfer_reset(&transfer, 0, NULL);
 transfer.type=LIBUSB_REQUEST_TYPE_STANDARD;
 transfer.recipient=LIBUSB_RECIPIENT_DEVICE;
 transfer.request=LIBUSB_REQUEST_SET_CONFIGURATION;
 transfer.value=0x01;
 transfer.index=0x00;

 if(usb_control(&dev, &transfer)!=0)
 {
  usb_close(&dev);
  return 2;
 }



 /*
  * End
  */

 if(midiOutputHandle)
  snd_rawmidi_close(midiOutputHandle);

 usb_transfer_reset(&transfer, 0, NULL);
 usb_close(&dev);
 return 0;
}

/*
 INIT
 ffff88005e334300 458923745 S Ci:2:013:0 s 80 06 0100 0000 0012 18 <
 ffff88005e334300 458923925 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff88005e334300 458923984 S Ci:2:013:0 s 80 06 0200 0000 0009 9 <
 ffff88005e334300 458924170 C Ci:2:013:0 0 9 = 0902a700 030100c0 00
 ffff88005e334300 458924192 S Ci:2:013:0 s 80 06 0200 0000 00a7 167 <
 ffff88005e334300 458924420 C Ci:2:013:0 0 167 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc4ff000 459221430 S Ci:2:013:0 s 80 06 0100 0000 0040 64 <
 ffff8801fc4ff000 459221550 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc4ff000 459415239 S Ci:2:013:0 s 80 06 0100 0000 0012 18 <
 ffff8801fc4ff000 459415696 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc4ff000 459418233 S Ci:2:013:0 s 80 06 0200 0000 00ff 255 <
 ffff8801fc4ff000 459418659 C Ci:2:013:0 0 167 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc4ff000 459420199 S Ci:2:013:0 s 80 06 0300 0000 00ff 255 <
 ffff8801fc4ff000 459420653 C Ci:2:013:0 0 4 = 04030904
 ffff8801fc4ff000 459422237 S Ci:2:013:0 s 80 06 0302 0409 00ff 255 <
 ffff8801fc4ff000 459422422 C Ci:2:013:0 0 12 = 0c035400 44002d00 31003500
 ffff8801fc4ff000 459431481 S Ci:2:013:0 s 80 06 0100 0000 0012 18 <
 ffff8801fc4ff000 459431925 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc4ff000 459434248 S Ci:2:013:0 s 80 06 0200 0000 0050 80 <
 ffff8801fc4ff000 459434422 C Ci:2:013:0 0 80 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc4ff000 459437275 S Ci:2:013:0 s 80 06 0200 0000 00a7 167 <
 ffff8801fc4ff000 459437545 C Ci:2:013:0 0 167 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc4ff000 459442609 S Co:2:013:0 s 01 0b 0000 0000 0000 0
 ffff8801fc4ff000 459442795 C Co:2:013:0 0 0
 ffff8801fc4ff000 459444406 S Co:2:013:0 s 01 0b 0000 0001 0000 0
 ffff8801fc4ff000 459444546 C Co:2:013:0 0 0
 ffff8801fc4ff000 459446220 S Co:2:013:0 s 01 0b 0000 0002 0000 0
 ffff8801fc4ff000 459446645 C Co:2:013:0 0 0
 ffff880171da1780 459499280 S Co:2:013:0 s 01 0b 0000 0002 0000 0
 ffff880171da1780 459499422 C Co:2:013:0 0 0
 ffff880171da1780 459501356 S Co:2:013:0 s 01 0b 0000 0000 0000 0
 ffff880171da1780 459501547 C Co:2:013:0 0 0
 ffff8800720ca000 459503515 S Co:2:013:0 s 01 0b 0000 0001 0000 0
 ffff8800720ca000 459503941 C Co:2:013:0 0 0
 ffff8800720ca000 459505607 S Bi:2:013:2 -115 512 <


 DEINIT
 ffff8800720ca000 459505607 S Bi:2:013:2 -115 512 <
 ffff8800720ca000 550499250 C Bi:2:013:2 -2 0
 ffff88020be5b9c0 550499458 S Co:2:001:0 s 23 03 0004 0004 0000 0
 ffff88020be5b9c0 550499467 C Co:2:001:0 0 0
 ffff88020be5b9c0 550550063 S Ci:2:001:0 s a3 00 0000 0004 0004 4 <
 ffff88020be5b9c0 550550299 C Ci:2:001:0 0 4 = 03051000
 ffff88020be5b9c0 550601153 S Co:2:001:0 s 23 01 0014 0004 0000 0
 ffff88020be5b9c0 550601163 C Co:2:001:0 0 0
 ffff88020be5b9c0 550601179 S Ci:2:000:0 s 80 06 0100 0000 0040 64 <
 ffff88020be5b9c0 550601256 C Ci:2:000:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc40bb40 550601276 S Co:2:001:0 s 23 03 0004 0004 0000 0
 ffff8801fc40bb40 550601282 C Co:2:001:0 0 0
 ffff8801fc40bb40 550652134 S Ci:2:001:0 s a3 00 0000 0004 0004 4 <
 ffff8801fc40bb40 550652374 C Ci:2:001:0 0 4 = 03051000
 ffff8801fc40bb40 550702980 S Co:2:001:0 s 23 01 0014 0004 0000 0
 ffff8801fc40bb40 550702990 C Co:2:001:0 0 0
 ffff8801fc40bb40 550702994 S Co:2:000:0 s 00 05 000d 0000 0000 0
 ffff8801fc40bb40 550703130 C Co:2:000:0 0 0
 ffff8801fc40bb40 550714892 S Ci:2:013:0 s 80 06 0100 0000 0012 18 <
 ffff8801fc40bb40 550715484 C Ci:2:013:0 0 18 = 12010002 ff00ff40 82055001 01000102 0001
 ffff8801fc40bc00 550715496 S Ci:2:013:0 s 80 06 0200 0000 00a7 167 <
 ffff8801fc40bc00 550715750 C Ci:2:013:0 0 167 = 0902a700 030100c0 00090400 0000ff02 02000624 f1010000 09040001 01ff0202
 ffff8801fc40bb40 550715758 S Co:2:013:0 s 00 09 0001 0000 0000 0
 ffff8801fc40bb40 550715878 C Co:2:013:0 0 0
*/
